@@ -4,7 +4,6 @@
 #include <sys/wait.h>
 
 #include <ctype.h>
-#include <curses.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <locale.h>
@@ -13,12 +12,15 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <term.h>
 #include <termios.h>
 #include <time.h>
 #include <unistd.h>
 #include <wchar.h>
 
+/* curses */
+#include <curses.h>
+#include <term.h>
+
 /* Allow to lazyload items when a file is specified? This saves memory but
    increases some latency when seeking items. It also causes issues if the
    feed is changed while having the UI open (and offsets are changed). */
@@ -43,9 +45,11 @@
 #define THEME_INPUT_LABEL()           do { attrmode(ATTR_REVERSE_ON); } while(0)
 #define THEME_INPUT_NORMAL()          do {                            } while(0)
 
-static char *plumber = "xdg-open"; /* environment variable: $SFEED_PLUMBER */
-static char *piper = "sfeed_content"; /* environment variable: $SFEED_PIPER */
-static char *yanker = "xclip -r"; /* environment variable: $SFEED_YANKER */
+static char *plumbercmd = "xdg-open"; /* env variable: $SFEED_PLUMBER */
+static char *pipercmd = "sfeed_content"; /* env variable: $SFEED_PIPER */
+static char *yankercmd = "xclip -r"; /* env variable: $SFEED_YANKER */
+static char *markreadcmd = "sfeed_markread read"; /* env variable: $SFEED_MARK_READ */
+static char *markunreadcmd = "sfeed_markread unread"; /* env variable: $SFEED_MARK_UNREAD */
 
 enum {
 	ATTR_RESET = 0,	ATTR_BOLD_ON = 1, ATTR_FAINT_ON = 2, ATTR_REVERSE_ON = 7
@@ -109,14 +113,6 @@ struct statusbar {
 
 /* /UI */
 
-struct feed {
-	char         *name;     /* feed name */
-	char         *path;     /* path to feed or NULL for stdin */
-	unsigned long totalnew; /* amount of new items per feed */
-	unsigned long total;    /* total items */
-	FILE *fp;               /* file pointer */
-};
-
 struct item {
 	char *link; /* separate link field (always loaded in case of urlfile) */
 	char *fields[FieldLast];
@@ -127,8 +123,19 @@ struct item {
 	off_t offset; /* line offset in file for lazyload */
 };
 
-#undef err
-void err(int, const char *, ...);
+struct items {
+	struct item *items;     /* array of items */
+	size_t len;             /* amount of items */
+	size_t cap;             /* available capacity */
+};
+
+struct feed {
+	char         *name;     /* feed name */
+	char         *path;     /* path to feed or NULL for stdin */
+	unsigned long totalnew; /* amount of new items per feed */
+	unsigned long total;    /* total items */
+	FILE *fp;               /* file pointer */
+};
 
 void alldirty(void);
 void cleanup(void);
@@ -170,8 +177,9 @@ ttywritef(const char *fmt, ...)
 	int n;
 
 	va_start(ap, fmt);
-	n = vdprintf(1, fmt, ap);
+	n = vfprintf(stdout, fmt, ap);
 	va_end(ap);
+	fflush(stdout);
 
 	return n;
 }
@@ -184,9 +192,9 @@ ttywrite(const char *s)
 	return write(1, s, strlen(s));
 }
 
-/* like BSD err(), but calls cleanup() and _exit(). */
+/* print to stderr, call cleanup() and _exit(). */
 void
-err(int code, const char *fmt, ...)
+die(const char *fmt, ...)
 {
 	va_list ap;
 	int saved_errno;
@@ -195,14 +203,15 @@ err(int code, const char *fmt, ...)
 	cleanup();
 
 	va_start(ap, fmt);
-	vdprintf(2, fmt, ap);
+	vfprintf(stderr, fmt, ap);
 	va_end(ap);
 
 	if (saved_errno)
-		dprintf(2, ": %s", strerror(saved_errno));
+		fprintf(stderr, ": %s", strerror(saved_errno));
+	fflush(stderr);
 	write(2, "\n", 1);
 
-	_exit(code);
+	_exit(1);
 }
 
 void *
@@ -211,7 +220,7 @@ erealloc(void *ptr, size_t size)
 	void *p;
 
 	if (!(p = realloc(ptr, size)))
-		err(1, "realloc");
+		die("realloc");
 	return p;
 }
 
@@ -221,7 +230,7 @@ ecalloc(size_t nmemb, size_t size)
 	void *p;
 
 	if (!(p = calloc(nmemb, size)))
-		err(1, "calloc");
+		die("calloc");
 	return p;
 }
 
@@ -231,7 +240,7 @@ estrdup(const char *s)
 	char *p;
 
 	if (!(p = strdup(s)))
-		err(1, "strdup");
+		die("strdup");
 	return p;
 }
 
@@ -390,7 +399,6 @@ updatetitle(void)
 void
 appmode(int on)
 {
-	/*ttywrite(on ? "\x1b[?1049h" : "\x1b[?1049l");*/ /* smcup, rmcup */
 	ttywrite(tparm(on ? enter_ca_mode : exit_ca_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 }
 
@@ -403,58 +411,63 @@ mousemode(int on)
 void
 cursormode(int on)
 {
-	/*ttywrite(on ? "\x1b[?25h" : "\x1b[?25l");*/ /* DECTCEM (in)Visible cursor */
 	ttywrite(tparm(on ? cursor_normal : cursor_invisible, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 }
 
-void
-cursorsave(void)
-{
-	/*ttywrite("\x1b""7");*/
-	ttywrite(tparm(save_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-}
-
-void
-cursorrestore(void)
-{
-	/*ttywrite("\x1b""8");*/
-	ttywrite(tparm(restore_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-}
-
 void
 cursormove(int x, int y)
 {
-	/*ttywritef("\x1b[%d;%dH", y + 1, x + 1);*/
 	ttywrite(tparm(cursor_address, y, x, 0, 0, 0, 0, 0, 0, 0));
 }
 
+void
+cursorsave(void)
+{
+	/* do not save the cursor if it won't be restored anyway */
+	if (cursor_invisible)
+		ttywrite(tparm(save_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+}
+
+void
+cursorrestore(void)
+{
+	/* if the cursor cannot be hidden then move to a consistent position */
+	if (cursor_invisible)
+		ttywrite(tparm(restore_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+	else
+		cursormove(0, 0);
+}
+
 void
 attrmode(int mode)
 {
-	char *p;
-
-	/*ttywritef("\x1b[%dm", mode);*/
 	switch (mode) {
-	case ATTR_RESET: p = exit_attribute_mode; break;
-	case ATTR_BOLD_ON: p = enter_bold_mode; break;
-	case ATTR_FAINT_ON: p = enter_dim_mode; break;
-	case ATTR_REVERSE_ON: p = enter_standout_mode; break;
-	default: return;
+	case ATTR_RESET:
+		ttywrite(tparm(exit_attribute_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+		break;
+	case ATTR_BOLD_ON:
+		ttywrite(tparm(enter_bold_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+		break;
+	case ATTR_FAINT_ON:
+		ttywrite(tparm(enter_dim_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+		break;
+	case ATTR_REVERSE_ON:
+		ttywrite(tparm(enter_reverse_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+		break;
+	default:
+		return;
 	}
-	ttywrite(tparm(p, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 }
 
 void
 cleareol(void)
 {
-	/*ttywrite("\x1b[K");*/
 	ttywrite(tparm(clr_eol, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 }
 
 void
 clearscreen(void)
 {
-	/*ttywrite("\x1b[H\x1b[2J");*/
 	ttywrite(tparm(clear_screen, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 }
 
@@ -499,11 +512,6 @@ win_update(struct win *w, int width, int height)
 void
 resizewin(void)
 {
-	/*struct winsize winsz;
-	if (ioctl(0, TIOCGWINSZ, &winsz) == -1)
-		err(1, "ioctl");
-	win_update(&win, winsz.ws_col, winsz.ws_row);*/
-
 	setupterm(NULL, 1, NULL);
 	/* termios globals are changed: `lines` and `columns` */
 	win_update(&win, columns, lines);
@@ -519,6 +527,8 @@ init(void)
 	tcgetattr(0, &tsave);
 	memcpy(&tcur, &tsave, sizeof(tcur));
 	tcur.c_lflag &= ~(ECHO|ICANON);
+	tcur.c_cc[VMIN] = 1;
+	tcur.c_cc[VTIME] = 0;
 	tcsetattr(0, TCSANOW, &tcur);
 
 	resizewin();
@@ -558,7 +568,7 @@ pipeitem(const char *cmd, struct item *item, int field, int wantoutput)
 
 	switch ((pid = fork())) {
 	case -1:
-		err(1, "fork");
+		die("fork");
 	case 0:
 		if (!wantoutput) {
 			dup2(devnullfd, 1);
@@ -567,7 +577,7 @@ pipeitem(const char *cmd, struct item *item, int field, int wantoutput)
 
 		errno = 0;
 		if (!(fp = popen(cmd, "w")))
-			err(1, "popen");
+			die("popen: %s", cmd);
 		if (field == -1) {
 			for (i = 0; i < FieldLast; i++) {
 				if (i)
@@ -598,7 +608,7 @@ forkexec(char *argv[])
 {
 	switch (fork()) {
 	case -1:
-		err(1, "fork");
+		die("fork");
 	case 0:
 		dup2(devnullfd, 1);
 		dup2(devnullfd, 2);
@@ -905,14 +915,14 @@ readch(void)
 		switch (select(1, &readfds, NULL, NULL, &tv)) {
 		case -1:
 			if (errno != EINTR)
-				err(1, "select");
+				die("select");
 			return -2; /* EINTR: like a signal */
 		case 0:
 			return -3; /* time-out */
 		}
 
 		switch (read(0, &b, 1)) {
-		case -1: err(1, "read");
+		case -1: die("read");
 		case 0: return EOF;
 		default: return (int)b;
 		}
@@ -998,7 +1008,9 @@ statusbar_draw(struct statusbar *s)
 	cursorsave();
 	cursormove(s->x, s->y);
 	THEME_STATUSBAR();
-	printpad(s->text, s->width);
+	/* terminals without xenl (eat newline glitch) mess up scrolling when
+	   using the last cell on the last line on the screen. */
+	printpad(s->text, s->width - (!eat_newline_glitch));
 	attrmode(ATTR_RESET);
 	cursorrestore();
 	s->dirty = 0;
@@ -1042,28 +1054,39 @@ linetoitem(char *line, struct item *item)
 	return 0;
 }
 
-int
-feed_getitems(struct feed *f, FILE *fp, struct item **items, size_t *nitems)
+void
+feed_items_free(struct items *items)
 {
-	struct item *item;
+	size_t i;
+
+	for (i = 0; i < items->len; i++) {
+		free(items->items[i].line);
+		free(items->items[i].link);
+	}
+	free(items->items);
+	items->items = NULL;
+	items->len = 0;
+	items->cap = 0;
+}
+
+int
+feed_items_get(struct feed *f, FILE *fp, struct items *itemsret)
+{
+	struct item *item, *items = NULL;
 	char *line = NULL;
-	size_t cap, i, linesize = 0;
+	size_t cap, i, linesize = 0, nitems;
 	ssize_t linelen;
 	off_t offset;
-	int ret = -1;
 
-	*items = NULL;
-	*nitems = 0;
-
-	cap = 0;
+	cap = nitems = 0;
 	offset = 0;
 	for (i = 0; ; i++) {
 		if (i + 1 >= cap) {
 			cap = cap ? cap * 2 : 16;
-			*items = erealloc(*items, cap * sizeof(struct item));
+			items = erealloc(items, cap * sizeof(struct item));
 		}
 		if ((linelen = getline(&line, &linesize, fp)) > 0) {
-			item = (*items) + i;
+			item = &items[i];
 
 			item->offset = offset;
 			offset += linelen;
@@ -1085,18 +1108,24 @@ feed_getitems(struct feed *f, FILE *fp, struct item **items, size_t *nitems)
 			linetoitem(estrdup(line), item);
 #endif
 
-			(*nitems)++;
+			nitems++;
 		}
 		if (ferror(fp))
 			goto err;
 		if (linelen <= 0 || feof(fp))
 			break;
 	}
-	ret = 0;
+	free(line);
+	itemsret->cap = cap;
+	itemsret->items = items;
+	itemsret->len = nitems;
+	return 0;
 
 err:
 	free(line);
-	return ret;
+	feed_items_free(itemsret);
+
+	return -1;
 }
 
 void
@@ -1125,27 +1154,21 @@ updatenewitems(struct feed *f)
 void
 feed_load(struct feed *f, FILE *fp)
 {
-	static struct item *items = NULL;
-	static size_t nitems = 0;
+	static struct items items;
 	struct pane *p;
 	size_t i;
 
-	for (i = 0; i < nitems; i++) {
-		free(items[i].line);
-		free(items[i].link);
-	}
-	free(items);
-
-	if (feed_getitems(f, fp, &items, &nitems) == -1)
-		err(1, "%s: %s", __func__, f->name);
+	feed_items_free(&items);
+	if (feed_items_get(f, fp, &items) == -1)
+		die("%s: %s", __func__, f->name);
 
 	p = &panes[PaneItems];
 	p->pos = 0;
-	p->nrows = nitems;
+	p->nrows = items.len;
 	free(p->rows);
-	p->rows = ecalloc(sizeof(p->rows[0]), nitems + 1);
-	for (i = 0; i < nitems; i++)
-		p->rows[i].data = &items[i]; /* do not use pane_row_get */
+	p->rows = ecalloc(sizeof(p->rows[0]), items.len + 1);
+	for (i = 0; i < items.len; i++)
+		p->rows[i].data = &(items.items[i]); /* do not use pane_row_get */
 
 	updatenewitems(f);
 
@@ -1201,7 +1224,7 @@ feeds_set(struct feed *f)
 
 	if (f && f->path) {
 		if (!f->fp && !(f->fp = fopen(f->path, "rb")))
-			err(1, "fopen: %s", f->path);
+			die("fopen: %s", f->path);
 	}
 
 	feed_setenv(f);
@@ -1216,7 +1239,7 @@ feeds_load(struct feed *feeds, size_t nfeeds)
 	size_t i;
 
 	if ((comparetime = time(NULL)) == -1)
-		err(1, "time");
+		die("time");
 	/* 1 day is old news */
 	comparetime -= 86400;
 
@@ -1226,10 +1249,10 @@ feeds_load(struct feed *feeds, size_t nfeeds)
 		if (f->path) {
 			if (f->fp) {
 				if (fseek(f->fp, 0, SEEK_SET))
-					err(1, "fseek: %s", f->path);
+					die("fseek: %s", f->path);
 			} else {
 				if (!(f->fp = fopen(f->path, "rb")))
-					err(1, "fopen: %s", f->path);
+					die("fopen: %s", f->path);
 			}
 		}
 		if (!f->fp) {
@@ -1271,15 +1294,13 @@ int
 getsidebarwidth(void)
 {
 	struct feed *feed;
-	static char text[1024];
 	int i, len, width = 0;
 
 	for (i = 0; i < nfeeds; i++) {
 		feed = &feeds[i];
 
-		snprintf(text, sizeof(text), "%s (%lu/%lu)",
-		         feed->name, feed->totalnew, feed->total);
-		len = colw(text);
+		len = snprintf(NULL, 0, " (%lu/%lu)", feed->totalnew, feed->total) +
+			colw(feed->name);
 		if (len > width)
 			width = len;
 
@@ -1439,7 +1460,7 @@ mousereport(int button, int release, int x, int y)
 					row = pane_row_get(p, p->pos);
 					item = (struct item *)row->data;
 					markread(p, p->pos, p->pos, 1);
-					forkexec((char *[]) { plumber, item->fields[FieldLink], NULL });
+					forkexec((char *[]) { plumbercmd, item->fields[FieldLink], NULL });
 				}
 			}
 			break;
@@ -1451,7 +1472,7 @@ mousereport(int button, int release, int x, int y)
 				row = pane_row_get(p, p->pos);
 				item = (struct item *)row->data;
 				markread(p, p->pos, p->pos, 1);
-				pipeitem(piper, item, -1, 1);
+				pipeitem(pipercmd, item, -1, 1);
 			}
 			break;
 		case 3: /* scroll up */
@@ -1510,7 +1531,7 @@ item_row_get(struct pane *p, off_t pos)
 	f = curfeed;
 	if (f && f->path && f->fp && !item->line) {
 		if (fseek(f->fp, item->offset, SEEK_SET))
-			err(1, "fseek: %s", f->path);
+			die("fseek: %s", f->path);
 		linelen = getline(&line, &linesize, f->fp);
 
 		if (linelen <= 0)
@@ -1565,24 +1586,18 @@ markread(struct pane *p, off_t from, off_t to, int isread)
 	if (!urlfile || !p->nrows)
 		return;
 
-	if (isread) {
-		if (!(cmd = getenv("SFEED_MARK_READ")))
-			cmd = "sfeed_markread read";
-	} else {
-		if (!(cmd = getenv("SFEED_MARK_UNREAD")))
-			cmd = "sfeed_markread unread";
-	}
+	cmd = isread ? markreadcmd : markunreadcmd;
 
 	switch ((pid = fork())) {
 	case -1:
-		err(1, "fork");
+		die("fork");
 	case 0:
 		dup2(devnullfd, 1);
 		dup2(devnullfd, 2);
 
 		errno = 0;
 		if (!(fp = popen(cmd, "w")))
-			err(1, "popen");
+			die("popen: %s", cmd);
 
 		for (i = from; i <= to; i++) {
 			row = &(p->rows[i]); /* use pane_row_get: no need for lazyload */
@@ -1681,17 +1696,21 @@ main(int argc, char *argv[])
 
 #ifdef __OpenBSD__
 	if (pledge("stdio rpath tty proc exec", NULL) == -1)
-		err(1, "pledge");
+		die("pledge");
 #endif
 
 	setlocale(LC_CTYPE, "");
 
 	if ((tmp = getenv("SFEED_PLUMBER")))
-		plumber = tmp;
+		plumbercmd = tmp;
 	if ((tmp = getenv("SFEED_PIPER")))
-		piper = tmp;
+		pipercmd = tmp;
 	if ((tmp = getenv("SFEED_YANKER")))
-		yanker = tmp;
+		yankercmd = tmp;
+	if ((tmp = getenv("SFEED_MARK_READ")))
+		markreadcmd = "sfeed_markread read";
+	if ((tmp = getenv("SFEED_MARK_UNREAD")))
+		markunreadcmd = "sfeed_markread unread";
 	urlfile = getenv("SFEED_URL_FILE");
 
 	panes[PaneFeeds].row_format = feed_row_format;
@@ -1707,7 +1726,7 @@ main(int argc, char *argv[])
 		f = &feeds[0];
 		f->name = "stdin";
 		if (!(f->fp = fdopen(0, "rb")))
-			err(1, "fdopen");
+			die("fdopen");
 	} else {
 		for (i = 1; i < argc; i++) {
 			f = &feeds[i - 1];
@@ -1723,9 +1742,10 @@ main(int argc, char *argv[])
 
 	if (!isatty(0)) {
 		if ((fd = open("/dev/tty", O_RDONLY)) == -1)
-			err(1, "open: /dev/tty");
+			die("open: /dev/tty");
 		if (dup2(fd, 0) == -1)
-			err(1, "dup2: /dev/tty");
+			die("dup2(%d, 0): /dev/tty -> stdin", fd);
+		close(fd);
 	}
 	if (argc == 1)
 		feeds[0].fp = NULL;
@@ -1739,7 +1759,7 @@ main(int argc, char *argv[])
 	}
 
 	if ((devnullfd = open("/dev/null", O_WRONLY)) == -1)
-		err(1, "open: /dev/null");
+		die("open: /dev/null");
 
 	updatesidebar(onlynew);
 	updatetitle();
@@ -1898,7 +1918,7 @@ nextpage:
 				p = &panes[selpane];
 				row = pane_row_get(p, p->pos);
 				item = (struct item *)row->data;
-				forkexec((char *[]) { plumber, item->fields[FieldEnclosure], NULL });
+				forkexec((char *[]) { plumbercmd, item->fields[FieldEnclosure], NULL });
 			}
 			break;
 		case 'm': /* toggle mouse mode */
@@ -1933,7 +1953,7 @@ nextpage:
 				row = pane_row_get(p, p->pos);
 				item = (struct item *)row->data;
 				markread(p, p->pos, p->pos, 1);
-				forkexec((char *[]) { plumber, item->fields[FieldLink], NULL });
+				forkexec((char *[]) { plumbercmd, item->fields[FieldLink], NULL });
 			}
 			break;
 		case 'c': /* items: pipe TSV line to program */
@@ -1946,11 +1966,11 @@ nextpage:
 				row = pane_row_get(p, p->pos);
 				item = (struct item *)row->data;
 				switch (ch) {
-				case 'y': pipeitem(yanker, item, FieldLink, 0); break;
-				case 'E': pipeitem(yanker, item, FieldEnclosure, 0); break;
+				case 'y': pipeitem(yankercmd, item, FieldLink, 0); break;
+				case 'E': pipeitem(yankercmd, item, FieldEnclosure, 0); break;
 				default:
 					markread(p, p->pos, p->pos, 1);
-					pipeitem(piper, item, -1, 1);
+					pipeitem(pipercmd, item, -1, 1);
 					break;
 				}
 			}
